public class Matrice {

    private double[][] M;
    private int nbreDeLignes;
    private int nbreDeColonnes;

    /**
     * constante en dessous de laquelle les nombres sont considérés comme nuls
     */
    static final double Epsilon = Math.pow(10, -8);

    /**
     * Constructeur
     *
     * @param T : tableau représentant une matrice
     */
    public Matrice(double[][] T) throws Exception {
        this.nbreDeLignes = T.length;
        this.nbreDeColonnes = T[0].length;
        M = new double[this.nbreDeLignes][this.nbreDeColonnes];
        for (int i = 0; i < this.nbreDeLignes; i++) {
            if (T[i].length != this.nbreDeColonnes) {
                throw new Exception("La ligne " + (i + 1) + " du tableau "
                        + "en argument est de longueur " + T[i].length
                        + " alors que la première ligne est de longeur "
                        + this.nbreDeColonnes);
            }
            for (int j = 0; j < this.nbreDeColonnes; j++) {
                M[i][j] = T[i][j];
            }
        }
    }

    /**
     * Conversion en string pour l'affichage, cette fonction est appelée
     * automatiquement lors de la conversion en String
     *
     * @return la matrice sous forme de plusieurs ligne
     */
    public String toString() {
        String resultat = "";
        for (int i = 0; i < this.nbreDeLignes; i++) {
            for (int j = 0; j < this.nbreDeColonnes; j++) {
                resultat = resultat + " " + M[i][j];
            }
            resultat = resultat + "\n";//passage à la ligne
        }
        return resultat;
    }

    /**
     * Echange 2 lignes
     *
     * @param i, k : lignes à échanger
     */
    void echangeLignes(int i, int k) {
        double[] ligneTemp = M[i];
        M[i] = M[k];
        M[k] = ligneTemp;
    }

    /**
     * Multilplication par un scalaire d'une ligne
     *
     * @param k : ligne
     * @param c : scalaire
     */
    void multiplicationLigne(int k, double c) {
        for (int j = 0; j < this.nbreDeColonnes; j++) {
            M[k][j] = c * M[k][j];
        }
    }

    /**
     * Enlève à une ligne une autre ligne multipliee par un scalaire
     *
     * @param i, k : lignes
     * @param c : scalaire
     */
    void enleveAuneLigneUneAutreLigne(int i, int k, double c) {
        for (int j = 0; j < this.nbreDeColonnes; j++) {
            M[i][j] = M[i][j] - c * M[k][j];
        }
    }

    /**
     * retourne la position du k-ième Pivot de Gauss génère une exception s'il
     * n'y a pas de pivot non nul
     *
     * @param k : etape
     * @return la position du pivot
     */
    int retournePivot(int k) throws Exception {
        int res = k;
        for (int i = k; i < this.nbreDeLignes; i++) {
            if (Math.abs(M[res][k]) < Math.abs(M[i][k])) {
                res = i;
            }
        }
        if (Math.abs(M[res][k]) < Epsilon) {
            throw new Exception("Il n'y a pas de pivot non nul à l'étape " + k);
        }
        return res;
    }

    /**
     * Enlève à une ligne la ligne du pivot multipliee par le scalaire adhoc
     *
     * @param i, k : lignes
     */
    void enleveAuneLigneUneAutreLigne(int i, int k) {
        double c = M[i][k] / M[k][k];
        M[i][k] = 0;
        for (int j = k + 1; j < this.nbreDeColonnes; j++) {
            M[i][j] = M[i][j] - c * M[k][j];
        }
    }

    /**
     * Triangularisation du système
     */
    void triangularisation() throws Exception {
        try {
            for (int k = 0; k < this.nbreDeLignes; k++) {
                int i = retournePivot(k);
                if (i != k) {
                    echangeLignes(i, k);
                }
                for (i = k + 1; i < this.nbreDeLignes; i++) {
                    this.enleveAuneLigneUneAutreLigne(i, k);
                }
            }
        } catch (Exception e) {
            throw new Exception("Ce système n'est pas de Cramer. " + e);
        }
    }

    /**
     * Solution d'un système de Cramer sur la sortie standard
     */
    public void solutionSystemeCramer() {
        if (this.nbreDeColonnes != (this.nbreDeLignes + 1)) {
            System.out.println("Cette matrice ne représente pas un système");
            return;
        }
        try {
            this.triangularisation();
            double[] sol = new double[this.nbreDeLignes];
            for (int k = this.nbreDeLignes - 1; k >= 0; k--) {
                sol[k] = M[k][nbreDeColonnes - 1];
                for (int j = k + 1; j < this.nbreDeLignes; j++) {
                    sol[k] = sol[k] - M[k][j] * sol[j];
                }
                sol[k] = sol[k] / M[k][k];
            }
            System.out.println("La solution de ce système est : ");
            for (int k = 0; k < this.nbreDeLignes; k++) {
                System.out.println("x_" + (k + 1) + "=" + sol[k]);
            }

        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * Copie du tableau representant la matrice
     *
     * @return la copie
     */
    double[][] copie() {
        double[][] res = new double[this.nbreDeLignes][this.nbreDeColonnes];
        for (int i = 0; i < this.nbreDeLignes; i++) {
            for (int j = 0; j < this.nbreDeColonnes; j++) {
                res[i][j] = M[i][j];
            }
        }
        return res;
    }

    /**
     * Déterminant de la matrice
     *
     * @return le déterminant
     */
    public double determinant() throws Exception {
        if (this.nbreDeColonnes != this.nbreDeLignes) {
            throw new Exception("La matrice n'est pas carrée.");
        }
        double det = 1;
        double[][] aux = this.copie();
        try {
            for (int k = 0; k < this.nbreDeLignes; k++) {
                int i = retournePivot(k);
                if (i != k) {
                    echangeLignes(i, k);
                    det = -det;
                }
                det = det * M[k][k];
                for (i = k + 1; i < this.nbreDeLignes; i++) {
                    this.enleveAuneLigneUneAutreLigne(i, k);
                }
            }
        } catch (Exception e) { //cas d'un pivot nul
            det = 0;
        }
        M = aux;
        return det;
    }

    /**
     * Retourne la matrice identité de taille n
     *
     * @param n : taille
     * @return matrice identité
     */
    public static Matrice identite(int n) throws Exception {
        double[][] M = new double[n][n];
        for (int i = 0; i < n; i++) {
            M[i][i] = 1;
        }
        return new Matrice(M);
    }

    /**
     * Retourne l'inverse de la matrice
     *
     * @return inverse de la matrice
     */
    public Matrice inverse() throws Exception {
        if (this.nbreDeColonnes != this.nbreDeLignes) {
            throw new Exception("La matrice n'est pas carrée.");
        }
        double[][] aux = this.copie();
        Matrice res = identite(this.nbreDeLignes);
        try {
            for (int k = 0; k < this.nbreDeLignes; k++) {
                int i = retournePivot(k);
                if (i != k) {
                    this.echangeLignes(i, k);
                    res.echangeLignes(i, k);
                }
                for (i = k + 1; i < this.nbreDeLignes; i++) {
                    double c = M[i][k] / M[k][k];
                    this.enleveAuneLigneUneAutreLigne(i, k, c);
                    res.enleveAuneLigneUneAutreLigne(i, k, c);
                }
            }
        } catch (Exception e) { //cas d'un pivot nul
            M = aux;
            throw new Exception("La matrice n'est pas inversible.");
        }
        for (int k = this.nbreDeLignes - 1; k >= 0; k--) {
            for (int i = 0; i < k; i++) {
                res.enleveAuneLigneUneAutreLigne(i, k, M[i][k] / M[k][k]);
            }
            res.multiplicationLigne(k, 1 / M[k][k]);
        }
        M = aux;
        return res;
    }
}
